package ca;

import ca.models.*;

public class Road {  
  // Read Only variables for display:

  // length of the road
  public int LENGTH = 300;
  // Maximum speed
  public int MAXSPEED;
  // the road as an array of speeds
  public int[] speed;
  // local density of vehicles on the road [veh/site]
  public double dens;
  // local mean velocity [sites/timestep]
  public double v;
  // local flow of vehicles [vehicles/time]
  public double flow;
  // global frequency of gaps
  public double[] gapfreq;
  // global frequency of velocities
  public double[] vfreq;

  // point of measurement
  private final int mp = LENGTH-1;
  // length of the measurement
  private final int ml = LENGTH/3;
  // dummy for infinity
  private final double DBL_INF = 99999.99;

  // number of vehicles on the road
  private int cars;
  // 1/number of vehicles
  private double invcars;
  // cell length
  private double deltaX;
  // simulator used
  private Simulator sim;

  private Road() {}

  public Road (double density, double deltaX, Simulator sim) {
    this.deltaX = deltaX;
    this.MAXSPEED = (int)(135.0/(deltaX*3.6)); // 135 km/h, dt=1s
    this.LENGTH = (int)(2250/deltaX); // 2250 meter (300 sites with dx=7.5m)

    setSimulator (sim);

    // the road as an array of speeds
    speed    = new int[LENGTH];
    // frequency of the gaps
    gapfreq  = new double[LENGTH];
    // frequency of the velocities
    vfreq    = new double[MAXSPEED+1];
    
    cars = 0;
    
    // clear road
    for (int i=0;i<LENGTH;i++) {
      speed[i] = -1;
    }

    // put vehicles on the road
    int cars_abs = (int)(density*LENGTH);
    invcars = 1.0/((cars_abs > 0)? cars_abs: 1.0/DBL_INF);
    while (cars < cars_abs) {
      int i = (int)(Math.random()*LENGTH);		
      if (speed[i] == -1) {
	speed[i] = 0;
	cars++;
      }
    } 
  }

  public void setSimulator (Simulator sim) {
    this.sim = sim;
    sim.init (MAXSPEED, LENGTH, deltaX);
  }


  private int updateSpeed(int speed, int gap, 
			  double prob_slowdown, double lambda) {
    int oldspeed = speed;
    
    speed = sim.updateSpeed (speed, gap, prob_slowdown, lambda);
    
    if (speed > gap) {
      // System.out.println ("crash! v: " + oldspeed + " gap: " + gap +
      //                     " -> v: " + speed); 
      speed = gap;
    } 
    return speed;
  }

  ////////////////////////////////////
  // update of the cellulare automaton

  public void update(double prob_slowdown, double prob_create, 
		     double lambda) {  
    // number of vehicles
    int car_local = (int)(prob_create*LENGTH);
    // the different between the actual and the requested number of vehicles
    int diff_cars = cars-car_local;
    
    for(int i=0;i<LENGTH;i++) {
      gapfreq[i]=0.0;
    }
    for(int i=0;i<=MAXSPEED;i++) {
      vfreq[i]=0.0;
    }
    
    if (car_local > LENGTH) {
      System.out.println ("Oh no - too much!\n");
      car_local = LENGTH;
    }
    if (car_local < 1) {
      System.out.println ("Oh no - too little!\n");
      car_local = 1;
    }

    // match the number of vehicles
    if (diff_cars > 0) {
      // number of vehivles, that have to be killed
      int kill_cars = 0;
      while (kill_cars < diff_cars) {
	int i = (int)(Math.random()*LENGTH);		
	if (speed[i] != -1) {
	  speed[i] = -1;
	  kill_cars++;
	}
      }
      cars = car_local;
      invcars  = 1.0/((cars > 0)? cars: 1.0/DBL_INF);
    } else if (diff_cars < 0) {
      // number of vehivles, that have to be created
      int new_cars = 0;
      while (new_cars < Math.abs(diff_cars))
	{
	  int i = (int)(Math.random()*LENGTH);		
	  if (speed[i] == -1) {
	    int gap = i;
	    while (speed[++gap%LENGTH] == -1);
	    gap-=(i+1);
	    speed[i] = Math.min(gap,MAXSPEED);
	    new_cars++;
	  }
	}
      cars = car_local;
      invcars  = 1.0/((cars > 0)? cars: 1.0/DBL_INF);
    }
    
    // where is the 1st vehicle
    int i = -1;
    while((++i < LENGTH) && (speed[i] == -1));
    // go on until reaching the end of the lane
    while (i < LENGTH) {
      // searching for the vehicle ahead
      int gap = i;
      while (speed[++gap%LENGTH] == -1);
      
      // dinstance between two following vehicles
      gap-=(i+1);
      gapfreq[gap]+=((invcars < DBL_INF)? invcars: 0);
      
      ////////////////
      // update rules

      speed[i] = updateSpeed (speed[i], gap, prob_slowdown, lambda);
      vfreq[speed[i]]+=((invcars < DBL_INF)? invcars: 0);

      // next vehicle is on site j
      i+=(gap+1);
    }


    // move the vehicles
    
    // where is the 1st vehicle
    car_local = 0;
    i = -1;
    while((++i < LENGTH) && (speed[i] == -1));
    // go on until reaching the end of the lane
    while (i < LENGTH) {
      car_local++;
      int inext = i+speed[i];
      int j = inext%LENGTH;
      // exchange the positions
      if (i != j) {
	speed[j] = speed[i];
	speed[i] = -1;
      }
      // searching for the vehicle ahead
      while (speed[++inext%LENGTH] == -1);
      i = inext;
    }
  }
  

  ////////////////////
  // local measurement

  public void measure() {
    int vsum=0;
    int rhoc=0;

    for(int i=mp;i>mp-ml;i--) {
      if (speed[i] >= 0) {
	vsum+=speed[i];
	rhoc++;
      }
    }
    v = (double)(vsum)/(double)((rhoc > 0) ? rhoc : 1 );
    dens =(double)(rhoc)/(double)(ml);
    flow = v*dens;
  }  

  
}  // End of "class Road"
