package ca;

import java.applet.*;
import java.awt.*;
import ca.models.*;

///////////////////////////////////////////////////////////
// Simulation of traffic flow with cellular atomatons
// enhanced by Kai Bolay
// based on work by B. Eisenblaetter and L. Neubert 1997
// which was based on work by Cay Horstmann 1996 

public class RoadApplet extends Applet {    
  // model Parameters
  private double slowdown = 0.01;
  private double density  = 0.30;
  private double lambda   = 0.77;

  // Labels of the two scrollbars
  private Label label_slowdown, label_density, label_lambda;

  private Button start, stop, clear;
  
  private RoadCanvas canvas;
  private Scrollbar sb_slowdown;
  private Scrollbar sb_density;
  private Scrollbar sb_lambda;
  private Choice model_choice, dx_choice;
  
  private Simulator[] sim = {new NagelSchreckenberg(),
			     new BarlovicSchreckenberg(),
			     new Helbing(),
			     // new KernerKonhaeuser(),
			     // new BandoSugiyama()
                             };
  private int type = 2;
  private int[] dx_choices = {250, 750};
  private int  deltaX = 1; // default choice = 7.5m

  private boolean FirstStep = true;

  // stupid bug in JDK 1.1
  private final static boolean sbBroken =
  (new Scrollbar(Scrollbar.HORIZONTAL,20,10,0,20).getValue() != 20);

  public void init() {  
    int i;
    // scrollbars of simulation speed, slowdown probability and global density
        
    Panel p = new Panel();
    p.setLayout(new GridLayout(4,4));

    p.add(new Label("Cell Length dx"));
    p.add(dx_choice = new Choice());
    for (i=0; i < dx_choices.length; i++) {
      dx_choice.addItem (String.valueOf (dx_choices[i]/100.0));
    }
    dx_choice.select (deltaX); // set default
    p.add(new Label("meter"));

    p.add(model_choice = new Choice());
    for (i=0; i < sim.length; i++) {
      model_choice.addItem (sim[i].getName());
    }
    model_choice.select (type); // set default

    p.add(new Label("Global Density"));
    i = (int)((100000*density)/dx_choices[deltaX]);
    sb_density = getSB (10, 120, i);
    p.add(sb_density);
    p.add(label_density = new Label((new Integer(i).toString())+" cars/km"));

    p.add(start = new Button("Start"));

    p.add(new Label("Slowdown Prob. P"));
    i = (int)(slowdown*100);
    sb_slowdown =  getSB (0, 100, i);
    p.add(sb_slowdown);
    p.add(label_slowdown = new Label((new Integer(i).toString())+"%"));

    p.add(stop = new Button("Stop"));
    
    p.add(new Label("Lambda"));
    i = (int)(lambda*100);
    sb_lambda = getSB (0, 100, i);
    p.add(sb_lambda);
    p.add(label_lambda = new Label((new Integer(i).toString())+"%"));

    p.add(clear = new Button("Clear Diagrams"));

    canvas = new RoadCanvas (density, slowdown, lambda,
			     dx_choices[deltaX]/100.0,  sim[type]);

    setLayout (new BorderLayout());
    add("North", p);
    add("Center", canvas);
    validate();
  }

  private Scrollbar getSB (int min, int max, int init) {
    final int inc = 1;
    return new Scrollbar(Scrollbar.HORIZONTAL, init, inc, min, 
			 max+(sbBroken?inc:0));
  }

  private void setPercentLabel (Label l, int v) {
    l.setText (String.valueOf (v)+"%");
  }
  
  public boolean handleEvent (Event evt) {
    switch (evt.id) {
    case Event.SCROLL_LINE_UP:
    case Event.SCROLL_LINE_DOWN:
    case Event.SCROLL_PAGE_UP:
    case Event.SCROLL_PAGE_DOWN:
    case Event.SCROLL_ABSOLUTE:
      if (evt.target == sb_density) {
	int v = sb_density.getValue();
	double dens = (v*dx_choices[deltaX])/100000.0;
	if (dens != density) {
	  density = dens;      
	  label_density.setText (String.valueOf (v)+" cars/km");
	}
      } else if (evt.target == sb_slowdown) {
	int v = sb_slowdown.getValue();
	double slow =  v*0.01;
	if (slow != slowdown) {
	  slowdown = slow;
	  setPercentLabel (label_slowdown, v);
	}
      } else if (evt.target == sb_lambda) {
	int v = sb_lambda.getValue();
	double lamb =  v*0.01;
	if (lamb != lambda) {
	  lambda = lamb;
	  setPercentLabel (label_lambda, v);
	}
      }
      canvas.newValues (density, slowdown, lambda); 
    }
    return super.handleEvent(evt);
  }

  public boolean action (Event evt, Object arg) {  
    int v;

    if (evt.target == start) {  
      start();
    } else if (evt.target == stop) {
      stop();
    } else if (evt.target == clear) {
      canvas.ClearDiagrams();
    } else if (evt.target == model_choice) {
      type = 0;
      while (!arg.equals (sim[type].getName())) type++;

      if ((sim[type] instanceof NagelSchreckenberg) ||
	  (sim[type] instanceof BarlovicSchreckenberg)) {
	lambda = 1.0;
	slowdown = 0.5;
      } else {
	lambda = 0.77;
	slowdown = 0.01;
      }

      v = (int)(slowdown*100);
      sb_slowdown.setValue(v);      
      setPercentLabel (label_slowdown, v);
      v = (int)(lambda*100);
      sb_lambda.setValue(v);      
      setPercentLabel (label_lambda, v);

      canvas.newValues (density, slowdown, lambda);

      canvas.freeway.setSimulator (sim[type]);
    } else if (evt.target == dx_choice) {
      canvas.stop(); // stop everything
      // remove old canvas
      remove (canvas);

      density /= dx_choices[deltaX]; // rescale density

      deltaX = 0;
      while (!arg.equals (String.valueOf (dx_choices[deltaX]/100.0))) deltaX++;

      density *= dx_choices[deltaX]; // rescale density
      
      // create, add and start new cc
      canvas = new RoadCanvas (density, slowdown, lambda, 
			       dx_choices[deltaX]/100.0, sim[type]);
      add("Center", canvas);
      validate();
      canvas.start();
    } else {
      return super.action(evt, arg);
    }
    return true;
  }
 
  /*
  public void paint() {
    System.out.println ("paint()");
  }
  */

  public void start() {  
    canvas.start();
  }
  
  public void stop() { 
    canvas.stop();
  }

  public void destroy() {
    stop();
  }
}
